/* smooth-mp.c, M. Welsh <mdw@cs.cornell.edu>
 * 
 * Fork off 4 threads to smooth an image. Uses libadt, 
 * gregk's smoothing routines, and Solaris 2.3 thread library.
 * Not very good since it has to crop and glue the input image;
 * this will be corrected soon.
 *
 */

#ifdef SVR4

#define _REENTRANT

#include <thread.h>

#include "image.h"
#include "vector.h"
#include "misc.h"
#include "smooth-gregk.h"

#define NUM_THRDS 4

static char rcsid[] = "";

static struct {
  float x, y;
} sm_offs[NUM_THRDS] = { 
 { 0.0, 0.0 },
 { 0.5, 0.0 },
 { 0.0, 0.5 },
 { 0.5, 0.5 },
};

struct sm_thr_data {
  GrayImage gim;
  FloatVector mask;
  FloatImage out;
  int thread_num;
};

static void *do_smooth_thread(void *tdata) {

    int in_x, in_y, wid, hgt;
    struct sm_thr_data *mydat;
    GrayImage inpiece;
    FloatImage outpiece;
    
    mydat = (struct sm_thr_data *)tdata;
     
    in_x = (sm_offs[mydat->thread_num].x) * imGetWidth(mydat->gim);
    in_y = (sm_offs[mydat->thread_num].y) * imGetHeight(mydat->gim);
    /* XXX For now assuming 4 threads... */
    wid = imGetWidth(mydat->gim) * 0.5;
    hgt = imGetHeight(mydat->gim) * 0.5;

    /* If in_y is nonzero this will break */
    inpiece = (GrayImage)imCropInto((GrayImage)mydat->gim,NULL,in_x,in_y,0,0,wid,hgt);
    /* XXX exit() from assert() not a good idea... */
    assert(inpiece != NULL); 

    outpiece = (FloatImage)smoothImage((GrayImage)inpiece,mydat->mask);  
    assert(outpiece != NULL); 

    imCropInto(outpiece,mydat->out,0,0,in_x,in_y,wid,hgt);  

    imFree(inpiece); /* imFree(outpiece); */

    thr_exit(NULL);

}

FloatImage smoothImage_mp(GrayImage gim, FloatVector mask) {

    thread_t mythreads[NUM_THRDS];
    FloatImage out;
    struct sm_thr_data *tdata[NUM_THRDS];
      
    int i;

    out = (FloatImage)imNew(IMAGE_FLOAT,imGetWidth(gim),imGetHeight(gim));
    assert(out != NULL);

    for (i = 0; i < NUM_THRDS; i++) {
      tdata[i] = (struct sm_thr_data *)malloc(sizeof(struct sm_thr_data));
      assert(tdata[i] != NULL);
      tdata[i]->gim = gim; tdata[i]->mask = mask;
      tdata[i]->out = out; tdata[i]->thread_num = i;
      thr_create(NULL,0,do_smooth_thread,(void *)tdata[i],
		 THR_NEW_LWP,&(mythreads[i]));
    }

    for (i = 0; i < NUM_THRDS; i++) {
      thr_join(mythreads[i],NULL,NULL);
    }

    return out;

}

#endif /* SVR4 */
